// ------------------------------------------------------------------------------------------------
#include "Base/Sphere.hpp"
#include "Base/Shared.hpp"
#include "Library/Random.hpp"

// ------------------------------------------------------------------------------------------------
namespace SqMod {

// ------------------------------------------------------------------------------------------------
const Sphere Sphere::NIL = Sphere();
const Sphere Sphere::MIN = Sphere(0.0);
const Sphere Sphere::MAX = Sphere(NumLimit< Sphere::Value >::Max);

// ------------------------------------------------------------------------------------------------
SQChar Sphere::Delim = ',';

// ------------------------------------------------------------------------------------------------
Sphere::Sphere()
    : pos(0.0), rad(0.0)
{
    /* ... */
}

// ------------------------------------------------------------------------------------------------
Sphere::Sphere(Value rv)
    : pos(0.0), rad(rv)
{
    /* ... */
}

// ------------------------------------------------------------------------------------------------
Sphere::Sphere(const Vector3 & pv, Value rv)
    : pos(pv), rad(rv)
{
    /* ... */
}

// ------------------------------------------------------------------------------------------------
Sphere::Sphere(Value xv, Value yv, Value zv, Value rv)
    : pos(xv, yv, zv), rad(rv)
{
    /* ... */
}

// ------------------------------------------------------------------------------------------------
Sphere::Sphere(const Sphere & o)
    : pos(o.pos), rad(o.rad)
{
    /* ... */
}

// ------------------------------------------------------------------------------------------------
Sphere::~Sphere()
{
    /* ... */
}

// ------------------------------------------------------------------------------------------------
Sphere & Sphere::operator = (const Sphere & o)
{
    pos = o.pos;
    rad = o.rad;
    return *this;
}

// ------------------------------------------------------------------------------------------------
Sphere & Sphere::operator = (Value r)
{
    rad = r;
    return *this;
}

Sphere & Sphere::operator = (const Vector3 & p)
{
    pos = p;
    return *this;
}

// ------------------------------------------------------------------------------------------------
Sphere & Sphere::operator += (const Sphere & s)
{
    pos += s.pos;
    rad += s.rad;
    return *this;
}

Sphere & Sphere::operator -= (const Sphere & s)
{
    pos -= s.pos;
    rad -= s.rad;
    return *this;
}

Sphere & Sphere::operator *= (const Sphere & s)
{
    pos *= s.pos;
    rad *= s.rad;
    return *this;
}

Sphere & Sphere::operator /= (const Sphere & s)
{
    pos /= s.pos;
    rad /= s.rad;
    return *this;
}

Sphere & Sphere::operator %= (const Sphere & s)
{
    pos %= s.pos;
    rad = fmod(rad, s.rad);

    return *this;
}

// ------------------------------------------------------------------------------------------------
Sphere & Sphere::operator += (Value r)
{
    rad += r;
    return *this;
}

Sphere & Sphere::operator -= (Value r)
{
    rad -= r;
    return *this;
}

Sphere & Sphere::operator *= (Value r)
{
    rad *= r;
    return *this;
}

Sphere & Sphere::operator /= (Value r)
{
    rad /= r;
    return *this;
}

Sphere & Sphere::operator %= (Value r)
{
    rad = fmod(rad, r);
    return *this;
}

// ------------------------------------------------------------------------------------------------
Sphere & Sphere::operator += (const Vector3 & p)
{
    pos += p;
    return *this;
}

Sphere & Sphere::operator -= (const Vector3 & p)
{
    pos -= p;
    return *this;
}

Sphere & Sphere::operator *= (const Vector3 & p)
{
    pos *= p;
    return *this;
}

Sphere & Sphere::operator /= (const Vector3 & p)
{
    pos /= p;
    return *this;
}

Sphere & Sphere::operator %= (const Vector3 & p)
{
    pos %= p;
    return *this;
}

// ------------------------------------------------------------------------------------------------
Sphere & Sphere::operator ++ ()
{
    ++pos;
    ++rad;
    return *this;
}

Sphere & Sphere::operator -- ()
{
    --pos;
    --rad;
    return *this;
}

// ------------------------------------------------------------------------------------------------
Sphere Sphere::operator ++ (int)
{
    Sphere state(*this);
    ++pos;
    ++rad;
    return state;
}

Sphere Sphere::operator -- (int)
{
    Sphere state(*this);
    --pos;
    --rad;
    return state;
}

// ------------------------------------------------------------------------------------------------
Sphere Sphere::operator + (const Sphere & s) const
{
    return Sphere(pos + s.pos, rad + s.rad);
}

Sphere Sphere::operator - (const Sphere & s) const
{
    return Sphere(pos - s.pos, rad - s.rad);
}

Sphere Sphere::operator * (const Sphere & s) const
{
    return Sphere(pos * s.pos, rad * s.rad);
}

Sphere Sphere::operator / (const Sphere & s) const
{
    return Sphere(pos / s.pos, rad / s.rad);
}

Sphere Sphere::operator % (const Sphere & s) const
{
    return Sphere(pos % s.pos, fmod(rad, s.rad));
}

// ------------------------------------------------------------------------------------------------
Sphere Sphere::operator + (Value r) const
{
    return Sphere(rad + r);
}

Sphere Sphere::operator - (Value r) const
{
    return Sphere(rad - r);
}

Sphere Sphere::operator * (Value r) const
{
    return Sphere(rad * r);
}

Sphere Sphere::operator / (Value r) const
{
    return Sphere(rad / r);
}

Sphere Sphere::operator % (Value r) const
{
    return Sphere(fmod(rad, r));
}

// ------------------------------------------------------------------------------------------------
Sphere Sphere::operator + (const Vector3 & p) const
{
    return Sphere(pos + p, rad);
}

Sphere Sphere::operator - (const Vector3 & p) const
{
    return Sphere(pos - p, rad);
}

Sphere Sphere::operator * (const Vector3 & p) const
{
    return Sphere(pos * p, rad);
}

Sphere Sphere::operator / (const Vector3 & p) const
{
    return Sphere(pos / p, rad);
}

Sphere Sphere::operator % (const Vector3 & p) const
{
    return Sphere(pos % p, rad);
}

// ------------------------------------------------------------------------------------------------
Sphere Sphere::operator + () const
{
    return Sphere(pos.Abs(), fabs(rad));
}

Sphere Sphere::operator - () const
{
    return Sphere(-pos, -rad);
}

// ------------------------------------------------------------------------------------------------
bool Sphere::operator == (const Sphere & s) const
{
    return EpsEq(rad, s.rad) && (pos == s.pos);
}

bool Sphere::operator != (const Sphere & s) const
{
    return !EpsEq(rad, s.rad) && (pos != s.pos);
}

bool Sphere::operator < (const Sphere & s) const
{
    return EpsLt(rad, s.rad) && (pos < s.pos);
}

bool Sphere::operator > (const Sphere & s) const
{
    return EpsGt(rad, s.rad) && (pos > s.pos);
}

bool Sphere::operator <= (const Sphere & s) const
{
    return EpsLtEq(rad, s.rad) && (pos <= s.pos);
}

bool Sphere::operator >= (const Sphere & s) const
{
    return EpsGtEq(rad, s.rad) && (pos >= s.pos);
}

// ------------------------------------------------------------------------------------------------
Int32 Sphere::Cmp(const Sphere & o) const
{
    if (*this == o)
        return 0;
    else if (*this > o)
        return 1;
    else
        return -1;
}

// ------------------------------------------------------------------------------------------------
CSStr Sphere::ToString() const
{
    return ToStringF("%f,%f,%f,%f", pos.x, pos.y, pos.z, rad);
}

// ------------------------------------------------------------------------------------------------
void Sphere::Set(Value nr)
{
    rad = nr;
}

void Sphere::Set(const Sphere & ns)
{
    pos = ns.pos;
    rad = ns.rad;
}

void Sphere::Set(const Vector3 & np)
{
    pos = np;
}

void Sphere::Set(const Vector3 & np, Value nr)
{
    pos = np;
    rad = nr;
}

// ------------------------------------------------------------------------------------------------
void Sphere::Set(Value nx, Value ny, Value nz)
{
    pos.Set(nx, ny, nz);
}

void Sphere::Set(Value nx, Value ny, Value nz, Value nr)
{
    pos.Set(nx, ny, nz);
    rad = nr;
}

// ------------------------------------------------------------------------------------------------
void Sphere::Set(CSStr values, SQChar delim)
{
    Set(GetSphere(values, delim));
}

// ------------------------------------------------------------------------------------------------
void Sphere::Generate()
{
    pos.Generate();
    rad = GetRandomFloat32();
}

void Sphere::Generate(Value min, Value max, bool r)
{
    if (EpsLt(max, min))
    {
        SqThrow("max value is lower than min value");
    }
    else if (r)
    {
        rad = GetRandomFloat32(min, max);
    }
    else
    {
        pos.Generate(min, max);
    }
}

void Sphere::Generate(Value xmin, Value xmax, Value ymin, Value ymax, Value zmin, Value zmax)
{
    pos.Generate(xmin, xmax, ymin, ymax, zmin, zmax);
}

void Sphere::Generate(Value xmin, Value xmax, Value ymin, Value ymax, Value zmin, Value zmax, Value rmin, Value rmax)
{
    if (EpsLt(rmax, rmin))
    {
        SqThrow("max value is lower than min value");
    }
    else
    {
        pos.Generate(xmin, xmax, ymin, ymax, zmin, zmax);
        rad = GetRandomFloat32(rmin, rmax);
    }
}

// ------------------------------------------------------------------------------------------------
Sphere Sphere::Abs() const
{
    return Sphere(pos.Abs(), fabs(rad));
}

// ================================================================================================
void Register_Sphere(HSQUIRRELVM vm)
{
    typedef Sphere::Value Val;

    RootTable(vm).Bind(_SC("Sphere"), Class< Sphere >(vm, _SC("Sphere"))
        /* Constructors */
        .Ctor()
        .Ctor< Val >()
        .Ctor< const Vector3 &, Val >()
        .Ctor< Val, Val, Val, Val >()
        /* Static Members */
        .SetStaticValue(_SC("Delim"), &Sphere::Delim)
        /* Member Variables */
        .Var(_SC("pos"), &Sphere::pos)
        .Var(_SC("rad"), &Sphere::rad)
        /* Properties */
        .Prop(_SC("abs"), &Sphere::Abs)
        /* Core Metamethods */
        .Func(_SC("_tostring"), &Sphere::ToString)
        .Func(_SC("_cmp"), &Sphere::Cmp)
        /* Metamethods */
        .Func<Sphere (Sphere::*)(const Sphere &) const>(_SC("_add"), &Sphere::operator +)
        .Func<Sphere (Sphere::*)(const Sphere &) const>(_SC("_sub"), &Sphere::operator -)
        .Func<Sphere (Sphere::*)(const Sphere &) const>(_SC("_mul"), &Sphere::operator *)
        .Func<Sphere (Sphere::*)(const Sphere &) const>(_SC("_div"), &Sphere::operator /)
        .Func<Sphere (Sphere::*)(const Sphere &) const>(_SC("_modulo"), &Sphere::operator %)
        .Func<Sphere (Sphere::*)(void) const>(_SC("_unm"), &Sphere::operator -)
        /* Setters */
        .Overload<void (Sphere::*)(const Sphere &)>(_SC("Set"), &Sphere::Set)
        .Overload<void (Sphere::*)(const Vector3 &, Val)>(_SC("Set"), &Sphere::Set)
        .Overload<void (Sphere::*)(Val, Val, Val, Val)>(_SC("Set"), &Sphere::Set)
        .Overload<void (Sphere::*)(Val)>(_SC("SetRad"), &Sphere::Set)
        .Overload<void (Sphere::*)(const Vector3 &)>(_SC("SetVec3"), &Sphere::Set)
        .Overload<void (Sphere::*)(Val, Val, Val)>(_SC("SetVec3"), &Sphere::Set)
        .Overload<void (Sphere::*)(CSStr, SQChar)>(_SC("SetStr"), &Sphere::Set)
        /* Utility Methods */
        .Func(_SC("Clear"), &Sphere::Clear)
        /* Operator Exposure */
        .Func<Sphere & (Sphere::*)(const Sphere &)>(_SC("opAddAssign"), &Sphere::operator +=)
        .Func<Sphere & (Sphere::*)(const Sphere &)>(_SC("opSubAssign"), &Sphere::operator -=)
        .Func<Sphere & (Sphere::*)(const Sphere &)>(_SC("opMulAssign"), &Sphere::operator *=)
        .Func<Sphere & (Sphere::*)(const Sphere &)>(_SC("opDivAssign"), &Sphere::operator /=)
        .Func<Sphere & (Sphere::*)(const Sphere &)>(_SC("opModAssign"), &Sphere::operator %=)

        .Func<Sphere & (Sphere::*)(Sphere::Value)>(_SC("opAddAssignR"), &Sphere::operator +=)
        .Func<Sphere & (Sphere::*)(Sphere::Value)>(_SC("opSubAssignR"), &Sphere::operator -=)
        .Func<Sphere & (Sphere::*)(Sphere::Value)>(_SC("opMulAssignR"), &Sphere::operator *=)
        .Func<Sphere & (Sphere::*)(Sphere::Value)>(_SC("opDivAssignR"), &Sphere::operator /=)
        .Func<Sphere & (Sphere::*)(Sphere::Value)>(_SC("opModAssignR"), &Sphere::operator %=)

        .Func<Sphere & (Sphere::*)(const Vector3 &)>(_SC("opAddAssignP"), &Sphere::operator +=)
        .Func<Sphere & (Sphere::*)(const Vector3 &)>(_SC("opSubAssignP"), &Sphere::operator -=)
        .Func<Sphere & (Sphere::*)(const Vector3 &)>(_SC("opMulAssignP"), &Sphere::operator *=)
        .Func<Sphere & (Sphere::*)(const Vector3 &)>(_SC("opDivAssignP"), &Sphere::operator /=)
        .Func<Sphere & (Sphere::*)(const Vector3 &)>(_SC("opModAssignP"), &Sphere::operator %=)

        .Func<Sphere & (Sphere::*)(void)>(_SC("opPreInc"), &Sphere::operator ++)
        .Func<Sphere & (Sphere::*)(void)>(_SC("opPreDec"), &Sphere::operator --)
        .Func<Sphere (Sphere::*)(int)>(_SC("opPostInc"), &Sphere::operator ++)
        .Func<Sphere (Sphere::*)(int)>(_SC("opPostDec"), &Sphere::operator --)

        .Func<Sphere (Sphere::*)(const Sphere &) const>(_SC("opAdd"), &Sphere::operator +)
        .Func<Sphere (Sphere::*)(const Sphere &) const>(_SC("opSub"), &Sphere::operator -)
        .Func<Sphere (Sphere::*)(const Sphere &) const>(_SC("opMul"), &Sphere::operator *)
        .Func<Sphere (Sphere::*)(const Sphere &) const>(_SC("opDiv"), &Sphere::operator /)
        .Func<Sphere (Sphere::*)(const Sphere &) const>(_SC("opMod"), &Sphere::operator %)

        .Func<Sphere (Sphere::*)(Sphere::Value) const>(_SC("opAddR"), &Sphere::operator +)
        .Func<Sphere (Sphere::*)(Sphere::Value) const>(_SC("opSubR"), &Sphere::operator -)
        .Func<Sphere (Sphere::*)(Sphere::Value) const>(_SC("opMulR"), &Sphere::operator *)
        .Func<Sphere (Sphere::*)(Sphere::Value) const>(_SC("opDivR"), &Sphere::operator /)
        .Func<Sphere (Sphere::*)(Sphere::Value) const>(_SC("opModR"), &Sphere::operator %)

        .Func<Sphere (Sphere::*)(const Vector3 &) const>(_SC("opAddP"), &Sphere::operator +)
        .Func<Sphere (Sphere::*)(const Vector3 &) const>(_SC("opSubP"), &Sphere::operator -)
        .Func<Sphere (Sphere::*)(const Vector3 &) const>(_SC("opMulP"), &Sphere::operator *)
        .Func<Sphere (Sphere::*)(const Vector3 &) const>(_SC("opDivP"), &Sphere::operator /)
        .Func<Sphere (Sphere::*)(const Vector3 &) const>(_SC("opModP"), &Sphere::operator %)

        .Func<Sphere (Sphere::*)(void) const>(_SC("opUnPlus"), &Sphere::operator +)
        .Func<Sphere (Sphere::*)(void) const>(_SC("opUnMinus"), &Sphere::operator -)

        .Func<bool (Sphere::*)(const Sphere &) const>(_SC("opEqual"), &Sphere::operator ==)
        .Func<bool (Sphere::*)(const Sphere &) const>(_SC("opNotEqual"), &Sphere::operator !=)
        .Func<bool (Sphere::*)(const Sphere &) const>(_SC("opLessThan"), &Sphere::operator <)
        .Func<bool (Sphere::*)(const Sphere &) const>(_SC("opGreaterThan"), &Sphere::operator >)
        .Func<bool (Sphere::*)(const Sphere &) const>(_SC("opLessEqual"), &Sphere::operator <=)
        .Func<bool (Sphere::*)(const Sphere &) const>(_SC("opGreaterEqual"), &Sphere::operator >=)
    );
}

} // Namespace:: SqMod
